/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "lock_manager.h"
#include "system/sm.h"
// #include "system/sm_meta.h"

LockDataId* LockManager::is_gap_compatible(LockDataId lock_rec_data_id,char* inrec_data,Transaction* txn){
    //先在迭代器里面找符合条件的fd
    for(auto it = Gap_lock_table_.begin(); it != Gap_lock_table_.end(); ++it){
        if(it->first.fd_==lock_rec_data_id.fd_){
            //证明表一样
            //循环请求队列
            for(auto req:it->second->request_queue_){
                if(req->txn_id_==txn->get_transaction_id()){
                    continue;
                }else{
                    char* rec_data;
                    std::unique_ptr<RmRecord> rec;
                    if(inrec_data!=nullptr){
                        rec_data=inrec_data;
                    }else{
                        //获取记录
                        RmFileHandle* fh_ = sm_manager_->fhs_.at(it->first.tab_name_).get();
                        rec=fh_->get_record_without_lock(lock_rec_data_id.rid_,nullptr);
                        rec_data=rec->data;
                    }
                    //判断是不是在范围内
                    int index_data_offset=0;
                    bool is_ingap=true;
                    for(auto index_col:it->first.index_meta_->cols){
                        char* lower_data=it->first.lower_+index_data_offset;
                        char* upper_data=it->first.upper_+index_data_offset;
                        char* data=rec_data+index_col.offset;
                        int low_num=ix_compare(lower_data,data,index_col.type,index_col.len);
                        int up_num=ix_compare(data,upper_data,index_col.type,index_col.len);
                        //没考虑开闭，开闭直接一起上
                        if(low_num<=0&&up_num<=0){
                            
                        }else{
                            is_ingap=false;
                            break;
                        }
                        index_data_offset=index_data_offset+index_col.len;
                    }
                    if(is_ingap){
                        //锁冲突处理
                        if(req->txn_start_ts<txn->get_start_ts()){
                            throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
                        }else{
                            //再等一等
                            const LockDataId* ptr = &it->first;
                            LockDataId *non_const_ptr = const_cast<LockDataId *>(ptr);
                            return non_const_ptr;
                        }
                    }else{
                        //如果不在这个间隙里面，并且对于这个请求队列来说所有的间隙都是相等的
                        break;
                    }
                }
            }
            
        }
    }
    return nullptr;
}

bool LockManager::isCompatible(LockMode existingLock, LockMode requestedLock) {
    int existingIndex = static_cast<int>(existingLock);
    int requestedIndex = static_cast<int>(requestedLock);
    return COMPATIBLITYMATRIX[existingIndex][requestedIndex];
}

/**
 * @description: 申请行级共享锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {Rid&} rid 加锁的目标记录ID 记录所在的表的fd
 * @param {int} tab_fd
 */
bool LockManager::lock_shared_on_record(Transaction* txn, const Rid& rid, int tab_fd) {
     //新的代码
    if(check(txn)==false){
        return false;
    }

    //新版上锁代码
    //构造上锁的lockid
    LockDataId lock_rec_data_id=LockDataId(tab_fd,rid,LockDataType::RECORD);
    //判断是否能上锁
    return general_lock(txn,lock_rec_data_id,LockMode::SHARED);
    //先把锁表锁住
    // std::unique_lock<std::mutex> lock{ latch_ };

    // //然后查看事务状态
    // if(txn->get_state()==TransactionState::SHRINKING){
    //     txn->set_state(TransactionState::ABORTED);
    //     throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
    // }
    // //设置事务的状态是增长
    // txn->set_state(TransactionState::GROWING);
    // //构造锁数据id
    // LockDataId lock_rec_data_id=LockDataId(tab_fd,rid,LockDataType::RECORD);
    // //首先查看是否位于间隙锁内
    // // while(true){
    // //     //首先查看是否位于间隙锁内
    // //     if(!is_gap_compatible(lock_rec_data_id)){
    // //         break;
    // //     }
    // //     Gap_lock_table_[lock_rec_data_id]->cv_.wait(lock);
    // // }
    // LockRequest* request = nullptr;
    // //根据id查看对应的锁请求
    // auto lockrequestqueue=lock_table_.find(lock_rec_data_id);
    // //如果请求队列是空
    // if(lockrequestqueue==lock_table_.end()){
    //     //构造一个新的请求队列
    //     std::unique_ptr<LockRequestQueue> lock_req_que=std::make_unique<LockRequestQueue>();
    //     lock_req_que->group_lock_mode_=GroupLockMode::S;
    //     //将新请求添加进去
    //     lock_req_que->request_queue_.emplace_back(txn->get_transaction_id(),LockMode::SHARED,false,txn->get_start_ts());
    //     //可能会有问题？      
    //     lock_table_[lock_rec_data_id]=std::move(lock_req_que);
        
    //     request = &(lock_table_[lock_rec_data_id]->request_queue_.back());
    // }else{
    //     //查看已有请求
    //     auto& req_que=lockrequestqueue->second->request_queue_;
    //     for(auto& it:req_que){
    //         if(it.txn_id_==txn->get_transaction_id()){
    //             //如果已经分配了
    //             if(it.granted_){
    //                 //查看已经分配了的锁的类型
    //                 if(it.lock_mode_>=LockMode::SHARED){
    //                     return true;
    //                 }else{
    //                     //现在只有读锁和写锁，好像不会有这种情况？
    //                 }
    //             }else{
    //                 assert(false);
    //                 it.lock_mode_=LockMode::SHARED>it.lock_mode_?LockMode::SHARED:it.lock_mode_;
    //                 request=&(it);
    //             }
    //             // break;
    //         }else{
    //             //锁冲突,其他事务占有锁并且是写锁，读锁可以一起上
    //             if(it.granted_&&(isCompatible(it.lock_mode_,LockMode::SHARED)==false)&&it.txn_start_ts<txn->get_start_ts()){
    //                 txn->set_state(TransactionState::ABORTED);
    //                 throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
    //             }else{

    //             }
    //         }
    //     }
    //     if(request==nullptr){
    //         //加入请求队列
    //         req_que.emplace_back(txn->get_transaction_id(),LockMode::SHARED,false,txn->get_start_ts());
    //         request=&req_que.back();
    //     }
    // }
    // //判断是否能上锁
    // while(true){//死循环，没上锁就等着
    //     auto& request_queue=lock_table_[lock_rec_data_id]->request_queue_;
    //     //遍历请求队列
    //     bool can_lock=true;
    //     for(auto it=request_queue.begin();it!=request_queue.end();it++){
    //         //如果已经为这个数据分配了锁
    //         if(it->granted_){
    //             //如果id相等，则不存在冲突
    //             if(it->txn_id_==txn->get_transaction_id()){
    //                 return true;
    //             }else if(it->txn_id_!=txn->get_transaction_id()){
    //                 if(!isCompatible(it->lock_mode_,request->lock_mode_)){
    //                     if(txn->get_start_ts()>it->txn_start_ts){//如果目前事务优先级低
    //                         txn->set_state(TransactionState::ABORTED);
    //                         throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
    //                     }else{
    //                         can_lock=false;
    //                     }
    //                 }else{
    //                     can_lock=true;
    //                 }
    //             }
    //         }else{//没有给事务分配锁，继续查看下一个
                
    //         }
    //     }
    //     if(can_lock){
    //         //std::cout<<"lock for:"<<request->txn_id_<<" s"<<std::endl;
    //         request->granted_=true;
    //         txn->get_lock_set()->insert(lock_rec_data_id);
    //         return true;
    //     }
    //     //等待变量等待别的线程释放
    //     lock_table_[lock_rec_data_id]->cv_.wait(lock);
    // }
    // return true;
}

/**
 * @description: 申请行级排他锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {Rid&} rid 加锁的目标记录ID
 * @param {int} tab_fd 记录所在的表的fd
 */
bool LockManager::lock_exclusive_on_record(Transaction* txn, const Rid& rid, int tab_fd,char* rec) {
    std::unique_lock<std::mutex> lock{ latch_ };
     //新的代码
    if(check(txn)==false){
        return false;
    }
    //新版上锁代码
    //构造上锁的lockid
    LockDataId lock_rec_data_id=LockDataId(tab_fd,rid,LockDataType::RECORD);
    while(true){
        //首先查看是否位于间隙锁内
        LockDataId* ptr=is_gap_compatible(lock_rec_data_id,rec,txn);
        if(ptr==nullptr){
            break;
        }else{
            Gap_lock_table_[*ptr]->cv_.wait(lock);
        }
    }
    //检查是否已经赋予锁
    //1.检查当前事务里面是否持有对这个数据的锁，锁升级是对于该事务而言
    //1.1持有对这个数据的锁
    if(txn->get_lock_set()->find(lock_rec_data_id)!=txn->get_lock_set()->end()){
        //1.2在锁表里面找该事务的锁,锁请求队列里面是已经赋予了吗？
        // auto lockrequestqueue=lock_table_.find(lock_table_data_id);
        // auto& req_que=lockrequestqueue;
        for(auto it=lock_table_[lock_rec_data_id].request_queue_.begin();it!=lock_table_[lock_rec_data_id].request_queue_.end();it++){
            if((*it)->txn_id_==txn->get_transaction_id()){
                assert((*it)->granted_==true);
                if((*it)->granted_==true){
                    if((*it)->lock_mode_==LockMode::EXLUCSIVE){
                        return true;
                    }else if((*it)->lock_mode_==LockMode::EXLUCSIVE){//如果是写锁，不需要锁升级
                        return true;
                    }
                    //其他情况都需要重新获取锁并且进行锁升级
                    else if(lock_table_[lock_rec_data_id].request_queue_.size()==1){
                        //如果这个数据的锁队列里面只有一个这个事务的请求
                        //并且请求已经授予
                        //直接进行锁请求修改(升级???)并且授予锁
                        (*it)->lock_mode_=LockMode::EXLUCSIVE;
                        lock_table_[lock_rec_data_id].group_lock_mode_=change_grouplockmode(LockMode::EXLUCSIVE);
                        return true;
                    }else{
                        //否则先释放这个请求然后重新获取
                        lock_table_[lock_rec_data_id].request_queue_.erase(it);
                        txn->get_lock_set()->erase(lock_rec_data_id);
                        lock_table_[lock_rec_data_id].group_lock_mode_ = change_grouplockmode(lock_table_[lock_rec_data_id].request_queue_.front()->lock_mode_);
                    }
                }
                break;
            }
        }
    }
    //构造本次请求
    auto lock_request = std::make_shared<LockRequest>(txn->get_transaction_id(), LockMode::EXLUCSIVE,false,txn->get_start_ts());
    //如果该数据项上面没有上锁队列，那么先构造一个上锁队列并且把请求放进去
    if(lock_table_.find(lock_rec_data_id)==lock_table_.end()){
        lock_table_[lock_rec_data_id].request_queue_.push_back(lock_request);
        lock_table_[lock_rec_data_id].group_lock_mode_=change_grouplockmode(LockMode::EXLUCSIVE);
    }else{
        while(!can_getlock(txn,lock_rec_data_id,LockMode::EXLUCSIVE)){
            if(txn->get_start_ts()>lock_table_[lock_rec_data_id].request_queue_.front()->txn_start_ts){
                txn->set_state(TransactionState::ABORTED);
                throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
            }
            lock_table_.at(lock_rec_data_id).cv_.wait(lock);
        }
        //后授予？？先授予？？
        lock_table_[lock_rec_data_id].request_queue_.push_back(lock_request);
        lock_table_[lock_rec_data_id].group_lock_mode_=change_grouplockmode(LockMode::EXLUCSIVE);
    }
    txn->get_lock_set()->insert(lock_rec_data_id); 
    lock_request->granted_ = true;
    return true;
    //新版代码
    //先把锁表锁住
    // std::unique_lock<std::mutex> lock{ latch_ };

    // //然后查看事务状态
    // if(txn->get_state()==TransactionState::SHRINKING){
    //     txn->set_state(TransactionState::ABORTED);
    //     throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
    // }
    // //设置事务的状态是增长
    // txn->set_state(TransactionState::GROWING);
    // //构造锁数据id
    // LockDataId lock_rec_data_id=LockDataId(tab_fd,rid,LockDataType::RECORD);
    
    // LockRequest* request = nullptr;
    // //根据id查看对应的锁请求
    // auto lockrequestqueue=lock_table_.find(lock_rec_data_id);
    // //如果请求队列是空
    // if(lockrequestqueue==lock_table_.end()){
    //     //构造一个新的请求队列
    //     std::unique_ptr<LockRequestQueue> lock_req_que=std::make_unique<LockRequestQueue>();
    //     lock_req_que->group_lock_mode_=GroupLockMode::X;
    //     //将新请求添加进去
    //     lock_req_que->request_queue_.emplace_back(txn->get_transaction_id(),LockMode::EXLUCSIVE,false,txn->get_start_ts());
    //     //可能会有问题？      
    //     lock_table_[lock_rec_data_id]=std::move(lock_req_que);
        
    //     request = &(lock_table_[lock_rec_data_id]->request_queue_.back());
    // }else{
    //     //查看已有请求
    //     auto& req_que=lockrequestqueue->second->request_queue_;
    //     lockrequestqueue->second->group_lock_mode_=GroupLockMode::X;        
    //     for(auto& it:req_que){
    //         if(it.txn_id_==txn->get_transaction_id()){
    //             //如果已经分配了
    //             if(it.granted_){
    //                 //查看已经分配了的锁的类型
    //                 if(it.lock_mode_==LockMode::EXLUCSIVE){
    //                     return true;
    //                 }else{
    //                     //如果当前的锁是s锁的话，升级请求
    //                     it.granted_=false;
    //                     it.lock_mode_=LockMode::EXLUCSIVE;
    //                     request=&(it);
                        
    //                 }
    //             }else{
    //                 it.lock_mode_=LockMode::EXLUCSIVE>it.lock_mode_?LockMode::EXLUCSIVE:it.lock_mode_;
    //                 request=&(it);
    //             }
    //         }else{
    //             //锁冲突,其他事务占有锁,并且该事务优先级更低，不需要请求了直接回滚
    //             if(it.granted_&&it.txn_start_ts<txn->get_start_ts()){
    //                 txn->set_state(TransactionState::ABORTED);
    //                 throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
    //             }else{

    //             }
    //         }
    //     }
    //     if(request==nullptr){
    //         //加入请求队列
    //         req_que.emplace_back(txn->get_transaction_id(),LockMode::EXLUCSIVE,false,txn->get_start_ts());
    //         request=&req_que.back();
    //     }
    // }
    // //首先查看是否位于间隙锁内
    // while(true){
    //     //首先查看是否位于间隙锁内
    //     LockDataId* ptr=is_gap_compatible(lock_rec_data_id,rec,txn);
    //     if(ptr==nullptr){
    //         break;
    //     }else{
    //         Gap_lock_table_[*ptr]->cv_.wait(lock);
    //     }
    // }
    // //判断是否能上锁
    // while(true){//死循环，没上锁就等着
    //     auto request_queue=lock_table_[lock_rec_data_id]->request_queue_;
    //     //遍历请求队列
    //     bool can_lock=true;
    //     for(auto it=request_queue.begin();it!=request_queue.end();it++){
    //         //如果已经为这个数据分配了锁
    //         if(it->granted_){
    //             //如果id相等，则不存在冲突
    //             if(it->txn_id_==txn->get_transaction_id()){
    //                 return true;
    //             }else if(it->txn_id_!=txn->get_transaction_id()){
    //                 if(!isCompatible(it->lock_mode_,request->lock_mode_)){
    //                     if(txn->get_start_ts()>it->txn_start_ts){//如果目前事务优先级低
    //                         txn->set_state(TransactionState::ABORTED);
    //                         throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
    //                     }else{
    //                         can_lock=false;
    //                     }
    //                 }else{
    //                     can_lock=true;
    //                 }
    //             }
    //         }else{//没有给事务分配锁，继续查看下一个
                
    //         }
    //     }
    //     if(can_lock){
    //         //std::cout<<"lock for:"<<request->txn_id_<<" x"<<std::endl;
    //         request->granted_=true;
    //         txn->get_lock_set()->insert(lock_rec_data_id);
    //         return true;
    //     }
    //     //等待变量等待别的线程释放
    //     lock_table_[lock_rec_data_id]->cv_.wait(lock);
    // }
    // return true;
}

//两阶段检查
bool LockManager::check(Transaction *txn){
    if(txn->get_state()==TransactionState::ABORTED||txn->get_state()==TransactionState::COMMITTED){
        return false;
    }else if(txn->get_state()==TransactionState::SHRINKING){
        txn->set_state(TransactionState::ABORTED);
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
        return false;
    }else if(txn->get_state()==TransactionState::DEFAULT){
        txn->set_state(TransactionState::GROWING);
        return true;
    }else if(txn->get_state()==TransactionState::GROWING){
        return true;
    }else{
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
    }
    return false;
}

//新版赋予锁的代码
bool LockManager::general_lock(Transaction* txn, LockDataId lock_table_data_id,LockMode lock_mode){
    std::unique_lock<std::mutex> lock(latch_);
    //检查是否已经赋予锁
    //1.检查当前事务里面是否持有对这个数据的锁，锁升级是对于该事务而言
    //1.1持有对这个数据的锁
    if(txn->get_lock_set()->find(lock_table_data_id)!=txn->get_lock_set()->end()){
        //1.2在锁表里面找该事务的锁,锁请求队列里面是已经赋予了吗？
        // auto lockrequestqueue=lock_table_.find(lock_table_data_id);
        // auto& req_que=lockrequestqueue;
        for(auto it=lock_table_[lock_table_data_id].request_queue_.begin();it!=lock_table_[lock_table_data_id].request_queue_.end();it++){
            if((*it)->txn_id_==txn->get_transaction_id()){
                assert((*it)->granted_==true);
                if((*it)->granted_==true){
                    if((*it)->lock_mode_==lock_mode){
                        return true;
                    }else if((*it)->lock_mode_==LockMode::EXLUCSIVE){//如果是写锁，不需要锁升级
                        return true;
                    }else if((*it)->lock_mode_==LockMode::S_IX&&
                    lock_table_data_id.type_==LockDataType::TABLE&&
                    lock_mode!=LockMode::EXLUCSIVE){
                        return true;
                    }
                    //其他情况都需要重新获取锁并且进行锁升级
                    else if(lock_table_[lock_table_data_id].request_queue_.size()==1){
                        //如果这个数据的锁队列里面只有一个这个事务的请求
                        //并且请求已经授予
                        //直接进行锁请求修改(升级???)并且授予锁
                        (*it)->lock_mode_=lock_mode;
                        lock_table_[lock_table_data_id].group_lock_mode_=change_grouplockmode(lock_mode);
                        return true;
                    }else{
                        //否则先释放这个请求然后重新获取
                        lock_table_[lock_table_data_id].request_queue_.erase(it);
                        txn->get_lock_set()->erase(lock_table_data_id);
                        lock_table_[lock_table_data_id].group_lock_mode_ = change_grouplockmode(lock_table_[lock_table_data_id].request_queue_.front()->lock_mode_);
                    }
                }
                break;
            }
        }
    }
    //构造本次请求
    auto lock_request = std::make_shared<LockRequest>(txn->get_transaction_id(), lock_mode,false,txn->get_start_ts());
    //如果该数据项上面没有上锁队列，那么先构造一个上锁队列并且把请求放进去
    if(lock_table_.find(lock_table_data_id)==lock_table_.end()){
        lock_table_[lock_table_data_id].request_queue_.push_back(lock_request);
        lock_table_[lock_table_data_id].group_lock_mode_=change_grouplockmode(lock_mode);
    }else{
        while(!can_getlock(txn,lock_table_data_id,lock_mode)){
            if(txn->get_start_ts()>lock_table_[lock_table_data_id].request_queue_.front()->txn_start_ts){
                txn->set_state(TransactionState::ABORTED);
                throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
            }
            lock_table_.at(lock_table_data_id).cv_.wait(lock);
        }
        //后授予？？先授予？？
        lock_table_[lock_table_data_id].request_queue_.push_back(lock_request);
        lock_table_[lock_table_data_id].group_lock_mode_=change_grouplockmode(lock_mode);
    }
    txn->get_lock_set()->insert(lock_table_data_id); 
    lock_request->granted_ = true;
    return true;
}

/**
 * @description: 申请表级读锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_shared_on_table(Transaction* txn, int tab_fd) {
    //先把锁表锁住
    // std::unique_lock<std::mutex> lock{ latch_ };

    //然后查看事务状态
    // if(txn->get_state()==TransactionState::SHRINKING){
    //     txn->set_state(TransactionState::ABORTED);
    //     throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
    // }
    if(check(txn)==false){
        return false;
    }

    //新版上锁代码
    //构造上锁的lockid
    LockDataId lock_table_data_id=LockDataId(tab_fd,LockDataType::TABLE);
    //判断是否能上锁
    return general_lock(txn,lock_table_data_id,LockMode::SHARED);
    //新版上锁代码
    // //设置事务的状态是增长
    // txn->set_state(TransactionState::GROWING);
    // //构造锁数据id
    // LockDataId lock_table_data_id=LockDataId(tab_fd,LockDataType::TABLE);
    // LockRequest* request = nullptr;
    // //根据id查看对应的锁请求
    // auto lockrequestqueue=lock_table_.find(lock_table_data_id);
    // //如果请求队列是空
    // if(lockrequestqueue==lock_table_.end()){
    //     //构造一个新的请求队列
    //     std::unique_ptr<LockRequestQueue> lock_req_que=std::make_unique<LockRequestQueue>();
    //     lock_req_que->group_lock_mode_=GroupLockMode::S;
    //     //将新请求添加进去
    //     lock_req_que->request_queue_.emplace_back(txn->get_transaction_id(),LockMode::SHARED,false,txn->get_start_ts());
    //     //可能会有问题？      
    //     lock_table_[lock_table_data_id]=std::move(lock_req_que);
        
    //     request = &(lock_table_[lock_table_data_id].request_queue_.back());
    // }else{
    //     //查看已有请求
    //     auto& req_que=lockrequestqueue->second->request_queue_;
    //     for(auto& it:req_que){
    //         if(it.txn_id_==txn->get_transaction_id()){
    //             //如果已经分配了
    //             if(it.granted_){
    //                 //查看已经分配了的锁的类型
    //                 if(it.lock_mode_>=LockMode::SHARED){
    //                     return true;
    //                 }else{
    //                     //否则需要进行锁升级
    //                     it.granted_=false;
    //                     //如果是is锁，则升级为s锁
    //                     if(it.lock_mode_==LockMode::INTENTION_SHARED){
    //                         it.lock_mode_=LockMode::SHARED;
    //                     }else{
    //                         it.lock_mode_=LockMode::S_IX;
    //                     }
    //                     request=&(it);
    //                 }
    //             }else{
    //                 assert(false);
    //                 it.lock_mode_=LockMode::SHARED>it.lock_mode_?LockMode::SHARED:it.lock_mode_;
    //                 request=&(it);
    //             }
    //             // break;
    //         }else{
    //             //锁冲突,其他事务占有锁并且是写锁，读锁可以一起上
    //             if(it.granted_&&(!isCompatible(it.lock_mode_,LockMode::SHARED))&&it.txn_start_ts<txn->get_start_ts()){
    //                 txn->set_state(TransactionState::ABORTED);
    //                 throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
    //             }else{

    //             }
    //         }
    //     }
    //     if(request==nullptr){
    //         //加入请求队列
    //         req_que.emplace_back(txn->get_transaction_id(),LockMode::SHARED,false,txn->get_start_ts());
    //         request=&req_que.back();
    //     }
    // }
    // //判断是否能上锁
    // while(true){//死循环，没上锁就等着
    //     auto& request_queue=lock_table_[lock_table_data_id].request_queue_;
    //     //遍历请求队列
    //     bool can_lock=true;
    //     for(auto it=request_queue.begin();it!=request_queue.end();it++){
    //         //如果已经为这个数据分配了锁
    //         if(it->granted_){
    //             //如果id相等，则不存在冲突
    //             if(it->txn_id_==txn->get_transaction_id()){
    //                 return true;
    //             }else if(it->txn_id_!=txn->get_transaction_id()){
    //                 if(!isCompatible(it->lock_mode_,request->lock_mode_)){
    //                     if(txn->get_start_ts()>it->txn_start_ts){//如果目前事务优先级低
    //                         txn->set_state(TransactionState::ABORTED);
    //                         throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
    //                     }else{
    //                         can_lock=false;
    //                     }
    //                 }else{
    //                     can_lock=true;
    //                 }
    //             }
    //         }else{//没有给事务分配锁，继续查看下一个
                
    //         }
    //     }
    //     if(can_lock){
    //         //std::cout<<"lock for:"<<request->txn_id_<<" s"<<std::endl;
    //         request->granted_=true;
    //         txn->get_lock_set()->insert(lock_table_data_id);
    //         return true;
    //     }
    //     //等待变量等待别的线程释放
    //     lock_table_[lock_table_data_id].cv_.wait(lock);
    // }
    // return true;
}

/**
 * @description: 申请表级写锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_exclusive_on_table(Transaction* txn, int tab_fd) {
    //新的代码
    if(check(txn)==false){
        return false;
    }

    //新版上锁代码
    //构造上锁的lockid
    LockDataId lock_table_data_id=LockDataId(tab_fd,LockDataType::TABLE);
    //判断是否能上锁
    return general_lock(txn,lock_table_data_id,LockMode::EXLUCSIVE);
    //新的代码


    // //先把锁表锁住
    // std::unique_lock<std::mutex> lock{ latch_ };

    // //然后查看事务状态
    // if(txn->get_state()==TransactionState::SHRINKING){
    //     txn->set_state(TransactionState::ABORTED);
    //     throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
    // }
    // //设置事务的状态是增长
    // txn->set_state(TransactionState::GROWING);
    // //构造锁数据id
    // LockDataId lock_table_data_id=LockDataId(tab_fd,LockDataType::TABLE);
    // LockRequest* request = nullptr;
    // //根据id查看对应的锁请求
    // auto lockrequestqueue=lock_table_.find(lock_table_data_id);
    // //如果请求队列是空
    // if(lockrequestqueue==lock_table_.end()){
    //     //构造一个新的请求队列
    //     std::unique_ptr<LockRequestQueue> lock_req_que=std::make_unique<LockRequestQueue>();
    //     lock_req_que->group_lock_mode_=GroupLockMode::X;
    //     //将新请求添加进去
    //     lock_req_que->request_queue_.emplace_back(txn->get_transaction_id(),LockMode::EXLUCSIVE,false,txn->get_start_ts());
    //     //可能会有问题？      
    //     lock_table_[lock_table_data_id]=std::move(lock_req_que);
        
    //     request = &(lock_table_[lock_table_data_id].request_queue_.back());
    // }else{
    //     //查看已有请求
    //     auto& req_que=lockrequestqueue->second->request_queue_;
    //     lockrequestqueue->second->group_lock_mode_=GroupLockMode::X;        
    //     for(auto& it:req_que){
    //         if(it.txn_id_==txn->get_transaction_id()){
    //             //如果已经分配了
    //             if(it.granted_){
    //                 //查看已经分配了的锁的类型,X锁是排他性最强的锁
    //                 if(it.lock_mode_==LockMode::EXLUCSIVE){
    //                     return true;
    //                 }else{
    //                     //如果当前的锁其他锁，升级请求
    //                     it.granted_=false;
    //                     it.lock_mode_=LockMode::EXLUCSIVE;
    //                     request=&(it);
                        
    //                 }
    //             }else{
    //                 it.lock_mode_=LockMode::EXLUCSIVE>it.lock_mode_?LockMode::EXLUCSIVE:it.lock_mode_;
    //                 request=&(it);
    //             }
    //         }else{
    //             //锁冲突,其他事务占有锁,并且该事务优先级更低，不需要请求了直接回滚
    //             if(it.granted_&&it.txn_start_ts<txn->get_start_ts()){
    //                 txn->set_state(TransactionState::ABORTED);
    //                 throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
    //             }else{

    //             }
    //         }
    //     }
    //     if(request==nullptr){
    //         //加入请求队列
    //         req_que.emplace_back(txn->get_transaction_id(),LockMode::EXLUCSIVE,false,txn->get_start_ts());
    //         request=&req_que.back();
    //     }
    // }
    // //判断是否能上锁
    // while(true){//死循环，没上锁就等着
    //     auto request_queue=lock_table_[lock_table_data_id].request_queue_;
    //     //遍历请求队列
    //     bool can_lock=true;
    //     for(auto it=request_queue.begin();it!=request_queue.end();it++){
    //         //如果已经为这个数据分配了锁
    //         if(it->granted_){
    //             //如果id相等，则不存在冲突
    //             if(it->txn_id_==txn->get_transaction_id()){
    //                 return true;
    //             }else if(it->txn_id_!=txn->get_transaction_id()){
    //                 if(!isCompatible(it->lock_mode_,request->lock_mode_)){
    //                     if(txn->get_start_ts()>it->txn_start_ts){//如果目前事务优先级低
    //                         txn->set_state(TransactionState::ABORTED);
    //                         throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
    //                     }else{
    //                         can_lock=false;
    //                     }
    //                 }else{
    //                     can_lock=true;
    //                 }
    //             }
    //         }else{//没有给事务分配锁，继续查看下一个
                
    //         }
    //     }
    //     if(can_lock){
    //         //std::cout<<"lock for:"<<request->txn_id_<<" x"<<std::endl;
    //         request->granted_=true;
    //         txn->get_lock_set()->insert(lock_table_data_id);
    //         return true;
    //     }
    //     //等待变量等待别的线程释放
    //     lock_table_[lock_table_data_id].cv_.wait(lock);
    // }
    // return true;
}

/**
 * @description: 申请表级意向读锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_IS_on_table(Transaction* txn, int tab_fd) {
    //新的代码
    if(check(txn)==false){
        return false;
    }

    //新版上锁代码
    //构造上锁的lockid
    LockDataId lock_table_data_id=LockDataId(tab_fd,LockDataType::TABLE);
    //判断是否能上锁
    return general_lock(txn,lock_table_data_id,LockMode::INTENTION_SHARED);
    // //先把锁表锁住
    // std::unique_lock<std::mutex> lock{ latch_ };

    // //然后查看事务状态
    // if(txn->get_state()==TransactionState::SHRINKING){
    //     txn->set_state(TransactionState::ABORTED);
    //     throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
    // }
    // //设置事务的状态是增长
    // txn->set_state(TransactionState::GROWING);
    // //构造锁数据id
    // LockDataId lock_table_data_id=LockDataId(tab_fd,LockDataType::TABLE);
    // LockRequest* request = nullptr;
    // //根据id查看对应的锁请求
    // auto lockrequestqueue=lock_table_.find(lock_table_data_id);
    // //如果请求队列是空
    // if(lockrequestqueue==lock_table_.end()){
    //     //构造一个新的请求队列
    //     std::unique_ptr<LockRequestQueue> lock_req_que=std::make_unique<LockRequestQueue>();
    //     lock_req_que->group_lock_mode_=GroupLockMode::IS;
    //     //将新请求添加进去
    //     lock_req_que->request_queue_.emplace_back(txn->get_transaction_id(),LockMode::INTENTION_SHARED,false,txn->get_start_ts());
    //     //可能会有问题？      
    //     lock_table_[lock_table_data_id]=std::move(lock_req_que);
        
    //     request = &(lock_table_[lock_table_data_id].request_queue_.back());
    // }else{
    //     //查看已有请求
    //     auto& req_que=lockrequestqueue->second->request_queue_;
    //     for(auto& it:req_que){
    //         if(it.txn_id_==txn->get_transaction_id()){
    //             //如果已经分配了
    //             if(it.granted_){
    //                 //查看已经分配了的锁的类型
    //                 if(it.lock_mode_>=LockMode::INTENTION_SHARED){
    //                     return true;
    //                 }else{
    //                     //由于is锁等级最低不需要锁升级
    //                     // //否则需要进行锁升级
    //                     // it.granted_=false;
    //                     // //如果是is锁，则升级为s锁
    //                     // if(it.lock_mode_==LockMode::INTENTION_SHARED){
    //                     //     it.lock_mode_=LockMode::SHARED;
    //                     // }else{
    //                     //     it.lock_mode_=LockMode::S_IX;
    //                     // }
    //                     // request=&(it);
    //                 }
    //             }else{
    //                 // assert(false);
    //                 it.lock_mode_=LockMode::INTENTION_SHARED>it.lock_mode_?LockMode::INTENTION_SHARED:it.lock_mode_;
    //                 request=&(it);
    //             }
    //             // break;
    //         }else{
    //             //锁冲突,其他事务占有写锁，其他锁可以一起上
    //             if(it.granted_&&(!isCompatible(it.lock_mode_,LockMode::INTENTION_SHARED))&&it.txn_start_ts<txn->get_start_ts()){
    //                 txn->set_state(TransactionState::ABORTED);
    //                 throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
    //             }else{

    //             }
    //         }
    //     }
    //     if(request==nullptr){
    //         //加入请求队列
    //         req_que.emplace_back(txn->get_transaction_id(),LockMode::INTENTION_SHARED,false,txn->get_start_ts());
    //         request=&req_que.back();
    //     }
    // }
    // //判断是否能上锁
    // while(true){//死循环，没上锁就等着
    //     auto& request_queue=lock_table_[lock_table_data_id].request_queue_;
    //     //遍历请求队列
    //     bool can_lock=true;
    //     for(auto it=request_queue.begin();it!=request_queue.end();it++){
    //         //如果已经为这个数据分配了锁
    //         if(it->granted_){
    //             //如果id相等，则不存在冲突
    //             if(it->txn_id_==txn->get_transaction_id()){
    //                 return true;
    //             }else if(it->txn_id_!=txn->get_transaction_id()){
    //                 if(!isCompatible(it->lock_mode_,request->lock_mode_)){
    //                     if(txn->get_start_ts()>it->txn_start_ts){//如果目前事务优先级低
    //                         txn->set_state(TransactionState::ABORTED);
    //                         throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
    //                     }else{
    //                         can_lock=false;
    //                     }
    //                 }else{
    //                     can_lock=true;
    //                 }
    //             }
    //         }else{//没有给事务分配锁，继续查看下一个
                
    //         }
    //     }
    //     if(can_lock){
    //         //std::cout<<"lock for:"<<request->txn_id_<<" is"<<std::endl;
    //         request->granted_=true;
    //         txn->get_lock_set()->insert(lock_table_data_id);
    //         return true;
    //     }
    //     //等待变量等待别的线程释放
    //     lock_table_[lock_table_data_id].cv_.wait(lock);
    // }
    return true;
}

/**
 * @description: 申请表级意向写锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_IX_on_table(Transaction* txn, int tab_fd) {
    //新的代码
    if(check(txn)==false){
        return false;
    }

    //新版上锁代码
    //构造上锁的lockid
    LockDataId lock_table_data_id=LockDataId(tab_fd,LockDataType::TABLE);
    //判断是否能上锁
    return general_lock(txn,lock_table_data_id,LockMode::INTENTION_EXCLUSIVE);
    // //先把锁表锁住
    // std::unique_lock<std::mutex> lock{ latch_ };

    // //然后查看事务状态
    // if(txn->get_state()==TransactionState::SHRINKING){
    //     txn->set_state(TransactionState::ABORTED);
    //     throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
    // }
    // //设置事务的状态是增长
    // txn->set_state(TransactionState::GROWING);
    // //构造锁数据id
    // LockDataId lock_table_data_id=LockDataId(tab_fd,LockDataType::TABLE);
    // LockRequest* request = nullptr;
    // //根据id查看对应的锁请求
    // auto lockrequestqueue=lock_table_.find(lock_table_data_id);
    // //如果请求队列是空
    // if(lockrequestqueue==lock_table_.end()){
    //     //构造一个新的请求队列
    //     std::unique_ptr<LockRequestQueue> lock_req_que=std::make_unique<LockRequestQueue>();
    //     lock_req_que->group_lock_mode_=GroupLockMode::IX;
    //     //将新请求添加进去
    //     lock_req_que->request_queue_.emplace_back(txn->get_transaction_id(),LockMode::INTENTION_EXCLUSIVE,false,txn->get_start_ts());
    //     //可能会有问题？      
    //     lock_table_[lock_table_data_id]=std::move(lock_req_que);
        
    //     request = &(lock_table_[lock_table_data_id].request_queue_.back());
    // }else{
    //     //查看已有请求
    //     auto& req_que=lockrequestqueue->second->request_queue_;
    //     for(auto& it:req_que){
    //         if(it.txn_id_==txn->get_transaction_id()){
    //             //如果已经分配了
    //             if(it.granted_){
    //                 //查看已经分配了的锁的类型
    //                 if(it.lock_mode_>=LockMode::INTENTION_EXCLUSIVE){
    //                     return true;
    //                 }else{
    //                     //由于is锁等级最低不需要锁升级
    //                     // //否则需要进行锁升级
    //                     if(it.lock_mode_==LockMode::INTENTION_SHARED){
    //                         it.granted_=false;
    //                         it.lock_mode_=LockMode::INTENTION_EXCLUSIVE;
    //                         request=&(it);
    //                     }else if(it.lock_mode_==LockMode::SHARED){
    //                         it.granted_=false;
    //                         it.lock_mode_=LockMode::S_IX;
    //                         request=&(it);
    //                     }
    //                 }
    //             }else{
    //                 // assert(false);
    //                 it.lock_mode_=LockMode::INTENTION_EXCLUSIVE>it.lock_mode_?LockMode::INTENTION_EXCLUSIVE:it.lock_mode_;
    //                 request=&(it);
    //             }
    //             // break;
    //         }else{
    //             //锁冲突,其他事务占有写锁，其他锁可以一起上
    //             if(it.granted_&&(!isCompatible(it.lock_mode_,LockMode::INTENTION_EXCLUSIVE))&&it.txn_start_ts<txn->get_start_ts()){
    //                 txn->set_state(TransactionState::ABORTED);
    //                 throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
    //             }else{

    //             }
    //         }
    //     }
    //     if(request==nullptr){
    //         //加入请求队列
    //         req_que.emplace_back(txn->get_transaction_id(),LockMode::INTENTION_EXCLUSIVE,false,txn->get_start_ts());
    //         request=&req_que.back();
    //     }
    // }
    // //判断是否能上锁
    // while(true){//死循环，没上锁就等着
    //     auto& request_queue=lock_table_[lock_table_data_id].request_queue_;
    //     //遍历请求队列
    //     bool can_lock=true;
    //     for(auto it=request_queue.begin();it!=request_queue.end();it++){
    //         //如果已经为这个数据分配了锁
    //         if(it->granted_){
    //             //如果id相等，则不存在冲突
    //             if(it->txn_id_==txn->get_transaction_id()){
    //                 return true;
    //             }else if(it->txn_id_!=txn->get_transaction_id()){
    //                 if(!isCompatible(it->lock_mode_,request->lock_mode_)){
    //                     if(txn->get_start_ts()>it->txn_start_ts){//如果目前事务优先级低
    //                         txn->set_state(TransactionState::ABORTED);
    //                         throw TransactionAbortException(txn->get_transaction_id(),AbortReason::DEADLOCK_PREVENTION);
    //                     }else{
    //                         can_lock=false;
    //                     }
    //                 }else{
    //                     can_lock=true;
    //                 }
    //             }
    //         }else{//没有给事务分配锁，继续查看下一个
                
    //         }
    //     }
    //     if(can_lock){
    //         //std::cout<<"lock for:"<<request->txn_id_<<" ix"<<std::endl;
    //         request->granted_=true;
    //         txn->get_lock_set()->insert(lock_table_data_id);
    //         return true;
    //     }
    //     //等待变量等待别的线程释放
    //     lock_table_[lock_table_data_id].cv_.wait(lock);
    // }
    return true;
}

//构建间隙锁
bool LockManager::lock_GAP_on_table(Transaction* txn, int tab_fd,IndexMeta* index_meta,char* upper,char* lower,int len,std::string tab_name)
{
    // //保护锁表并发
    std::unique_lock<std::mutex> lock{ latch_ };

    // //然后查看事务状态
    // if(txn->get_state()==TransactionState::SHRINKING){
    //     txn->set_state(TransactionState::ABORTED);
    //     throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
    // }
    // //设置事务的状态是增长
    // txn->set_state(TransactionState::GROWING);
    if(check(txn)==false){
        return false;
    }
    // //构造锁数据id
    LockDataId lock_gap_data_id=LockDataId(tab_fd,upper,lower,len,LockDataType::GAP,tab_name);
    lock_gap_data_id.index_meta_=new IndexMeta;
    lock_gap_data_id.index_meta_->col_num=index_meta->col_num;
    lock_gap_data_id.index_meta_->col_tot_len=index_meta->col_tot_len;
    lock_gap_data_id.index_meta_->cols=index_meta->cols;
    lock_gap_data_id.index_meta_->tab_name=index_meta->tab_name;
    // LockRequest* request = nullptr;
    // //根据id查看对应的锁请求
    auto lockrequestqueue=Gap_lock_table_.find(lock_gap_data_id);
    auto lock_request = std::make_shared<LockRequest>(txn->get_transaction_id(),LockMode::SHARED,false,txn->get_start_ts());
    //如果请求队列是空
    if(lockrequestqueue==Gap_lock_table_.end()){
        // //构造一个新的请求队列
        std::unique_ptr<LockRequestQueue> lock_req_que=std::make_unique<LockRequestQueue>();
        // lock_req_que->group_lock_mode_=GroupLockMode::S;
        // //将新请求添加进去
        // lock_req_que->request_queue_.emplace_back(txn->get_transaction_id(),LockMode::SHARED,true,txn->get_start_ts());
        // //可能会有问题？      
        Gap_lock_table_[lock_gap_data_id]=std::move(lock_req_que);
        
        // request = &(Gap_lock_table_[lock_gap_data_id]->request_queue_.back());
        Gap_lock_table_[lock_gap_data_id]->request_queue_.push_back(lock_request);
    }else{
        //查看已有请求
        auto& req_que=lockrequestqueue->second->request_queue_;
        for(auto& it:req_que){
            if(it->txn_id_==txn->get_transaction_id()){
                //如果已经分配了
                if(it->granted_){
                    return true;
                }
            }else{

            }
        }
        // if(request==nullptr){
        //     //加入请求队列
        //     // req_que.emplace_back(txn->get_transaction_id(),LockMode::SHARED,false,txn->get_start_ts());
        //     // request=&req_que.back();
        // }
        Gap_lock_table_[lock_gap_data_id]->request_queue_.push_back(lock_request);
    }
    //由于间隙锁都是读锁，因此都可以直接上锁
    //std::cout<<"lock for:"<<txn->get_transaction_id()<<" GAP"<<std::endl;
    lock_request->granted_=true;
    txn->get_lock_set()->insert(lock_gap_data_id);
    return true;
}

/**
 * @description: 释放锁
 * @return {bool} 返回解锁是否成功
 * @param {Transaction*} txn 要释放锁的事务对象指针
 * @param {LockDataId} lock_data_id 要释放的锁ID
 */
bool LockManager::unlock(Transaction* txn, LockDataId lock_data_id) {
    //保护锁表并发
    std::scoped_lock lock{latch_};
    //修改事务状态
    txn->set_state(TransactionState::SHRINKING);
    //如果锁表中不存在该对象，则直接返回
    if(!lock_table_.count(lock_data_id)&&!Gap_lock_table_.count(lock_data_id)){
        return false;
    }
    if(lock_table_.count(lock_data_id))
    {
        //删除锁请求队列中所出现的锁请求
        auto &request_queue=lock_table_[lock_data_id];
        bool can_erase=true;
        auto request=request_queue.request_queue_.begin();
        while(request!=request_queue.request_queue_.end()){
            if((*request)->txn_id_==txn->get_transaction_id()){
                //一个事务只能有一个锁
                if(lock_table_[lock_data_id].request_queue_.size()==1){
                    lock_table_[lock_data_id].group_lock_mode_=GroupLockMode::NON_LOCK;
                }else{
                    lock_table_[lock_data_id].group_lock_mode_=change_grouplockmode(lock_table_[lock_data_id].request_queue_.front()->lock_mode_);
                }
                request=request_queue.request_queue_.erase(request);
                break;
            }
            else{
                request++;
            }
            // else{
            //     can_erase=false;
            //     request++;
            // }
        }
        //释放当前锁id对应的资源
        // if(can_erase)
        // {
        //     lock_table_.erase(lock_data_id);
        // }
        // else{
        //     lock_table_[lock_data_id]->cv_.notify_all();
        // }
        lock_table_[lock_data_id].cv_.notify_all();
    }else{
        //删除锁请求队列中所出现的锁请求
        auto &request_queue=Gap_lock_table_[lock_data_id];
        bool can_erase=true;
        auto request=request_queue->request_queue_.begin();
        while(request!=request_queue->request_queue_.end()){
            if((*request)->txn_id_==txn->get_transaction_id()){
                request=request_queue->request_queue_.erase(request);
            }else{
                can_erase=false;
                request++;
            }
        }
        //释放当前锁id对应的资源
        // if(can_erase)
        // {
        //     Gap_lock_table_.erase(lock_data_id);
        // }
        // else{
        //     Gap_lock_table_[lock_data_id]->cv_.notify_all();
        // }
        Gap_lock_table_[lock_data_id]->cv_.notify_all();
    }
    
    return true;
}